@@ -10,6 +10,7 @@ n - sequence
 d - dimension
 l - logits (text)
 i, j - sequence (row, col)
+p - positions
 """
 
 import os
@@ -19,12 +20,14 @@ from functools import partial
 from typing import NamedTuple, Callable, Literal
 
 import torch
-from torch import nn, Tensor, tensor
+from torch import nn, Tensor, tensor, is_tensor
 import torch.nn.functional as F
 from torch.nn import Module, ModuleList, Linear
 from torch.nn.utils.rnn import pad_sequence
 from torch.utils._pytree import tree_map
 
+from torchdiffeq import odeint
+
 import einx
 from einops import rearrange, repeat, reduce, einsum, pack
 from einops.layers.torch import Rearrange
@@ -34,6 +37,7 @@ from ema_pytorch import EMA
 from rotary_embedding_torch import RotaryEmbedding, apply_rotary_emb
 
 from tqdm import tqdm
+from loguru import logger
 
 pad_sequence = partial(pad_sequence, batch_first = True)
 
@@ -107,7 +111,7 @@ def cast_tuple(t, length = 1):
     return t if isinstance(t, tuple) else ((t,) * length)
 
 def tree_map_tensor(sample, fn: Callable):
-    return tree_map(lambda t: t if not torch.is_tensor(t) else fn(t), sample)
+    return tree_map(lambda t: t if not is_tensor(t) else fn(t), sample)
 
 def eval_decorator(fn):
     def inner(self, *args, **kwargs):
@@ -161,7 +165,7 @@ def print_modality_sample(
         else:
             output.append(('modality', sample.shape))
 
-    print(output)
+    logger.info(output)
 
 # character based tokenizer
 
@@ -386,7 +390,60 @@ def min_p_filter(logits, min_p = 0.1):
     limit = min_p * max_probs
     return torch.where(probs < limit, float('-inf'), logits)
 
-from torchdiffeq import odeint
+# MLP parameterized N-dimensional positions
+
+class MLPAxialPositions(Module):
+    def __init__(
+        self,
+        *,
+        num_dimensions, # 2 for images, 3 for video, etc etc
+        dim,
+        expand_factor = 2.
+    ):
+        super().__init__()
+        self.num_dimensions = num_dimensions
+        dim_hidden = int(dim * expand_factor)
+
+        self.mlp = nn.Sequential(
+            nn.Linear(num_dimensions, dim),
+            nn.SiLU(),
+            nn.Linear(dim, dim_hidden),
+            nn.SiLU(),
+            nn.Linear(dim_hidden, dim)
+        )
+
+        # tensor typing
+
+        self._d = dim
+
+    @property
+    def device(self):
+        return next(self.parameters()).device
+
+    @typecheck
+    def forward(
+        self,
+        modality_shape: Int['p'] | torch.Size,
+        flatten_dims = False
+    ) -> Float['... {self._d}']:
+
+        if isinstance(modality_shape, torch.Size):
+            modality_shape = tensor(modality_shape)
+
+        modality_shape = modality_shape.to(self.device)
+
+        assert len(modality_shape) == self.num_dimensions
+        dimensions = modality_shape.tolist()
+
+        grid = torch.meshgrid([torch.arange(dim_len, device = self.device) for dim_len in dimensions], indexing = 'ij')
+        axial_positions = torch.stack(grid, dim = -1)
+
+        pos_emb = self.mlp(axial_positions.float())
+
+        if flatten_dims:
+            pos_emb = rearrange(pos_emb, '... d -> (...) d')
+
+        return pos_emb
 
 # random fourier embedding
 
@@ -843,7 +900,7 @@ class Transformer(Module):
             x = x[..., cache_length:, :]
             cond = cond[..., cache_length:, :]
 
-            if torch.is_tensor(is_any_modality):
+            if is_tensor(is_any_modality):
                 is_any_modality = is_any_modality[..., cache_length:]
 
         # adaptive layernorm kwargs, which handles text and modality tokens differently
@@ -920,11 +977,13 @@ class Transfusion(Module):
         transformer: dict | Transformer,
         dim_latent: int | tuple[int, ...] | None = None,
         channel_first_latent = False,
+        add_pos_emb: bool | tuple[bool, ...] = False,
         modality_encoder: Module | tuple[Module, ...] | None = None,
         modality_decoder: Module | tuple[Module, ...] | None = None,
         modality_token_transform: tuple[ModalityTokenTransform, ...] | ModalityTokenTransform | None = None,
         modality_default_shape: tuple[int, ...] | tuple[tuple[int, ...], ...] | None = None,
-        modality_validate_num_dim: int | tuple[int, ...] | None = None,
+        fallback_to_default_shape_if_invalid = False,
+        modality_num_dim: int | tuple[int, ...] | None = None,
         to_modality_shape_fn: Callable | tuple[Callable, ...] = default_to_modality_shape_fn,
         ignore_index = -1,
         flow_loss_weight = 1.,
@@ -969,8 +1028,30 @@ class Transfusion(Module):
 
         # specifying the number of dimensions for the modality, which will be hard validated
 
-        self.modality_validate_num_dim = cast_tuple(modality_validate_num_dim, self.num_modalities)
-        assert len(self.modality_validate_num_dim) == self.num_modalities
+        self.modality_num_dim = cast_tuple(modality_num_dim, self.num_modalities)
+        assert len(self.modality_num_dim) == self.num_modalities
+
+        # whether to add an extra axial positional embedding per modality
+
+        self.add_pos_emb = cast_tuple(add_pos_emb, self.num_modalities)
+        assert len(self.add_pos_emb) == self.num_modalities
+
+        self.pos_emb_mlp = ModuleList([])
+
+        for modality_add_pos_emb, modality_ndim in zip(self.add_pos_emb, self.modality_num_dim):
+
+            if not modality_add_pos_emb:
+                self.pos_emb_mlp.append(None)
+                continue
+
+            assert exists(modality_ndim), '`modality_num_dim` must be set if you wish to automatically inject axial positional embeddings'
+
+            pos_generating_mlp = MLPAxialPositions(
+                dim = dim,
+                num_dimensions = modality_ndim,
+            )
+
+            self.pos_emb_mlp.append(pos_generating_mlp)
 
         # modality encoders and decoders
 
@@ -993,6 +1074,8 @@ class Transfusion(Module):
 
         assert len(self.modality_default_shape) == self.num_modalities
 
+        self.fallback_to_default_shape_if_invalid = fallback_to_default_shape_if_invalid
+
         # entire "sentence" start and end id
 
         num_text_special_ids = 2
@@ -1119,7 +1202,7 @@ class Transfusion(Module):
             maybe_meta_tensor = get_tokens_since_rightmost_id(seq, self.meta_id)
 
             default_shape = self.modality_default_shape[curr_modality_id]
-            maybe_modality_validate_num_dim = self.modality_validate_num_dim[curr_modality_id]
+            maybe_modality_num_dim = self.modality_num_dim[curr_modality_id]
 
             meta_str_to_modality_shape = self.to_modality_shape_fn[curr_modality_id]
 
@@ -1136,8 +1219,14 @@ class Transfusion(Module):
 
             modality_shape = default(modality_shape, default_shape)
 
+            if self.fallback_to_default_shape_if_invalid:
+
+                if exists(maybe_modality_num_dim) and len(modality_shape) != maybe_modality_num_dim:
+                    logger.warning(f'invalid modality shape {modality_shape} for modality {curr_modality_id}. falling back to default shape {default_shape}')
+                    modality_shape = default_shape
+
             assert exists(modality_shape), f'language model did not produce a proper modality shape for modality type {curr_modality_id} - please set a fallback shape with `modality_default_shape`'
-            assert not exists(maybe_modality_validate_num_dim) or maybe_modality_validate_num_dim == len(modality_shape), f'expected modality type {curr_modality_id} to have {maybe_modality_validate_num_dim} dimensions but language model produced a shape of {modality_shape}'
+            assert not exists(maybe_modality_num_dim) or maybe_modality_num_dim == len(modality_shape), f'expected modality type {curr_modality_id} to have {maybe_modality_num_dim} dimensions but language model produced a shape of {modality_shape}'
 
             is_decoding_text = False
 
@@ -1216,6 +1305,8 @@ class Transfusion(Module):
                         step_times = rearrange(step_times, ' -> 1 1') # batch size of 1
                         step_times = F.pad(step_times, (num_past_modalities, 0), value = 1.) # past decoded modalities receive a time conditioning of 1.
 
+                        denoised = denoised.reshape(*modality_shape, latent_dim)
+
                         embeds, new_kv_cache = self.forward(
                             [[*modality_sample, (curr_modality_id, denoised)]],
                             times = step_times,
@@ -1356,6 +1447,8 @@ class Transfusion(Module):
         return_loss = True
     ) -> Float['']:
 
+        shape = modalities.shape
+
         if self.num_modalities > 1:
             assert exists(modality_type), '`modality_type` must be explicitly passed in on forward when training on greater than 1 modality'
 
@@ -1365,12 +1458,32 @@ class Transfusion(Module):
         latent_to_model_fn = self.latent_to_model_projs[modality_type]
         model_to_flow_pred_fn = self.model_to_latent_preds[modality_type]
 
+        # grab the shape of the modality, for maybe axial pos emb
+
+        add_pos_emb = self.add_pos_emb[modality_type]
+        maybe_pos_emb_mlp = self.pos_emb_mlp[modality_type]
+        modality_num_dim = self.modality_num_dim[modality_type]
+
+        if add_pos_emb:
+            assert exists(modality_num_dim), f'modality_num_dim must be set for modality {modality_type} if further injecting axial positional embedding'
+
+            if self.channel_first_latent:
+                _, _, *axial_dims = shape
+            else:
+                _, *axial_dims, _ = shape
+
+            assert len(axial_dims) == modality_num_dim, f'received modalities of ndim {len(axial_dims)} but expected {modality_num_dim}'
+
+        # maybe transform
+
         tokens = transform(modalities)
 
         # maybe channel first
 
         if self.channel_first_latent:
-            tokens = rearrange(tokens, 'b d ... -> b (...) d')
+            tokens = rearrange(tokens, 'b d ... -> b ... d')
+
+        tokens = rearrange(tokens, 'b ... d -> b (...) d')
 
         # rotary
 
@@ -1391,10 +1504,17 @@ class Transfusion(Module):
 
         flow = tokens - noise
 
-        # attention
-
         noised_tokens = latent_to_model_fn(noised_tokens)
 
+        # maybe add axial pos emb
+
+        if add_pos_emb:
+            axial_pos_emb = maybe_pos_emb_mlp(tensor(axial_dims), flatten_dims = True)
+
+            noised_tokens = noised_tokens + axial_pos_emb
+
+        # attention
+
         embed = self.transformer(
             noised_tokens,
             times = times,
@@ -1441,8 +1561,8 @@ class Transfusion(Module):
     ):
         is_decoding = exists(decoding_text_or_modality)
 
-        is_text_only = torch.is_tensor(modalities) and modalities.dtype in (torch.int, torch.long)
-        is_modality_only = torch.is_tensor(modalities) and modalities.dtype == torch.float
+        is_text_only = is_tensor(modalities) and modalities.dtype in (torch.int, torch.long)
+        is_modality_only = is_tensor(modalities) and modalities.dtype == torch.float
 
         # handle ema model being passed in for velocity consistency loss
 
@@ -1489,15 +1609,22 @@ class Transfusion(Module):
                 prepend(modality, tensor_([self.sos_id]))
                 modality.append(tensor_([self.eos_id]))
 
+        # need axial pos emb
+
+        need_axial_pos_emb = any(self.add_pos_emb)
+
         # process list of text and modalities interspersed with one another
 
         modality_positions = []
         modality_tokens = []
+        modality_pos_emb = []
+
         text = []
 
         for batch_modalities in modalities:
             batch_modality_positions = []
             batch_modality_tokens = []
+            batch_modality_pos_emb = []
             batch_text = []
             offset = 0
 
@@ -1505,7 +1632,7 @@ class Transfusion(Module):
                 # if non-text modality detected and not given as a tuple
                 # cast to (int, Tensor) where int is defaulted to type 0 (convenience for one modality)
 
-                if torch.is_tensor(modality) and modality.dtype == torch.float:
+                if is_tensor(modality) and modality.dtype == torch.float:
                     modality = (0, modality)
 
                 is_text = not isinstance(modality, tuple)
@@ -1548,6 +1675,11 @@ class Transfusion(Module):
 
                     batch_text.append(modality_tensor)
                     offset += text_length
+
+                    if need_axial_pos_emb:
+                        zeros = torch.zeros((text_length, self.dim), device = device)
+                        batch_modality_pos_emb.append(zeros)
+
                     continue
 
                 # otherwise handle a modality
@@ -1566,6 +1698,8 @@ class Transfusion(Module):
                 modality_shape_str = join([*map(str, modality_shape_tuple)], ',')
                 modality_meta_info = self.char_tokenizer(modality_shape_str, device = device)
 
+                prec_meta_tag_length = len(modality_meta_info) + 2
+
                 text_tensor = torch.cat((
                     tensor_([self.meta_id]),
                     modality_meta_info,
@@ -1583,12 +1717,37 @@ class Transfusion(Module):
 
                 offset += modality_length + 2 + len(modality_meta_info) + 1 # +2 due to [som] and [eom] - then account for meta start id and modality shape information (or eventually any meta information about modality)
 
+                # handle axial positional embedding
+
+                if need_axial_pos_emb:
+
+                    modality_pos_mlp = self.pos_emb_mlp[modality_type]
+
+                    if exists(modality_pos_mlp):
+                        pos_emb = modality_pos_mlp(tensor(modality_shape_tuple), flatten_dims= True)
+
+                        pos_emb = F.pad(pos_emb, (0, 0, prec_meta_tag_length, 1), value = 0.)
+                    else:
+                        pos_emb = torch.zeros(text_tensor.shape[0], self.dim, device = device)
+
+                    batch_modality_pos_emb.append(pos_emb)
+
             text.append(torch.cat(batch_text))
+
+            if need_axial_pos_emb:
+                modality_pos_emb.append(torch.cat(batch_modality_pos_emb, dim = -2))
+
             modality_tokens.append(batch_modality_tokens)
             modality_positions.append(batch_modality_positions)
 
         text = pad_sequence(text, padding_value = -1)
 
+        if need_axial_pos_emb:
+            modality_pos_emb = pad_sequence(modality_pos_emb, padding_value = 0.)
+
+            if return_loss:
+                modality_pos_emb = modality_pos_emb[:, :-1]
+
         # if returning loss, split text for next token prediction
 
         if return_loss:
@@ -1613,7 +1772,7 @@ class Transfusion(Module):
 
         # embed the list of modality tokens into a sequence of Float['b n d'] at right offsets and lengths as dictated by modalities info tensor
 
-        if torch.is_tensor(modality_tokens):
+        if is_tensor(modality_tokens):
             modality_tokens = [modality_tokens]
 
         # embed the modality tokens into one Tensor if not given as one
@@ -1647,6 +1806,13 @@ class Transfusion(Module):
 
         times_per_token = einsum(is_modalities.float(), times, 'b t m n, b m -> b t n')
 
+        # if needs velocity matching, make sure times are in the range of 0 - (1. - <velocity consistency delta time>)
+
+        if need_velocity_matching:
+            times_per_token = times_per_token * (1. - velocity_consistency_delta_time)
+
+        # noise only if returning loss
+
         if return_loss:
             noised_modality_tokens = []
             flows = []
@@ -1676,6 +1842,11 @@ class Transfusion(Module):
 
         modality_tokens = sum(modality_tokens)
 
+        # maybe add the axial positional embedding
+
+        if need_axial_pos_emb:
+            modality_tokens = modality_tokens + modality_pos_emb
+
         # intersperse the modalities with the text for the joint transformer + flow system
 
         tokens = einx.where('b n, b n d, b n d', is_any_modality, modality_tokens, text_tokens)
